// ReSharper disable CppTooWideScope
// ReSharper disable CppDefaultCaseNotHandledInSwitchStatement
// ReSharper disable CppClangTidyHicppMultiwayPathsCovered
#include "MarchingCubesGL.h"

#include <iostream>

namespace
{
    // Light Colors:
    constexpr GLfloat LightColorAmbient[]  = {0.50, 0.50, 0.50, 1.00};
    constexpr GLfloat LightColorDiffuse[]  = {0.75, 0.75, 0.75, 1.00};
    constexpr GLfloat LightColorSpecular[] = {1.00, 1.00, 1.00, 1.00};

    // Cooper: http://devernay.free.fr/cours/opengl/materials.html
    constexpr GLfloat ColorFrontAmbient[]  = {0.19125f, 0.0735f, 0.0225f, 1.00};
    constexpr GLfloat ColorFrontDiffuse[]  = {0.7038f, 0.27048f, 0.0828f, 1.00};
    constexpr GLfloat ColorFrontSpecular[] = {0.256777f, 0.137622f, 0.086014f, 1.00};
    constexpr GLfloat ColorFrontShininess  = 0.1f * 128.0;

    // Emerald: http://devernay.free.fr/cours/opengl/materials.html
    constexpr GLfloat ColorBackAmbient[] = {0.0215f, 0.1745f, 0.0215f, 1.00};
    constexpr GLfloat ColorBackDiffuse[] = {0.07568f, 0.61424f, 0.07568f, 1.00};

    // Data structure for Potential Surface (Sapmle7)
    struct Dodecahedron
    {
        double S           = 0.23;    // Scaling
        double Data[20][3] = {
            {               -S * sqrt(1 + 2.0 / sqrt(5)),                  +S * 0.0,           +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            {               +S * sqrt(1 + 2.0 / sqrt(5)),                  +S * 0.0,           -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            { -S * sqrt(1 / 8.0 + 1.0 / (8.0 * sqrt(5))), -S * 0.25 * (3 + sqrt(5)),           +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            { -S * sqrt(1 / 8.0 + 1.0 / (8.0 * sqrt(5))), +S * 0.25 * (3 + sqrt(5)),           +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            {+S * sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5))), -S * 0.25 * (1 + sqrt(5)),           +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},

            {+S * sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5))), +S * 0.25 * (1 + sqrt(5)),           +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            {      -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0), -S * 0.25 * (1 + sqrt(5)),      +S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {      -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0), +S * 0.25 * (1 + sqrt(5)),      +S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {         -S * 0.5 * sqrt(1 + 2.0 / sqrt(5)),                  -S * 0.5,      -S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {         -S * 0.5 * sqrt(1 + 2.0 / sqrt(5)),                  +S * 0.5,      -S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},

            { +S * sqrt(1 / 4.0 + 1.0 / (2.0 * sqrt(5))),                  -S * 0.5, +S* std::sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            { +S * sqrt(1 / 4.0 + 1.0 / (2.0 * sqrt(5))),                  +S * 0.5, +S* std::sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {            +S * sqrt((5 + sqrt(5)) / 10.0),                  +S * 0.0, -S* std::sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {-S * sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5))), -S * 0.25 * (1 + sqrt(5)),           -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            {-S * sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5))), +S * 0.25 * (1 + sqrt(5)),           -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},

            { -S * sqrt(1 / 2.0 + 1.0 / (2.0 * sqrt(5))),                  +S * 0.0,      +S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {      +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0), -S * 0.25 * (1 + sqrt(5)),      -S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            {      +S * 0.5 * sqrt((5 - sqrt(5)) / 10.0), +S * 0.25 * (1 + sqrt(5)),      -S* sqrt(5 / 8.0 + 11.0 / (8.0 * sqrt(5)))},
            { +S * sqrt(1 / 8.0 + 1.0 / (8.0 * sqrt(5))), -S * 0.25 * (3 + sqrt(5)),           -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)},
            { +S * sqrt(1 / 8.0 + 1.0 / (8.0 * sqrt(5))), +S * 0.25 * (3 + sqrt(5)),           -S * 0.5 * sqrt((5 - sqrt(5)) / 10.0)}
        };
    };

    // VertexOffset lists the positions, relative to vertex0, of each of the 8 vertices of a cube
    const GLfloat VertexOffset[8][3] = {
        {0.0, 0.0, 0.0},
        {1.0, 0.0, 0.0},
        {1.0, 1.0, 0.0},
        {0.0, 1.0, 0.0},
        {0.0, 0.0, 1.0},
        {1.0, 0.0, 1.0},
        {1.0, 1.0, 1.0},
        {0.0, 1.0, 1.0}
    };

    // EdgeConnection lists the index of the endpoint vertices for each of the 12 edges of the cube
    const GLint EdgeConnection[12][2] = {
        {0, 1},
        {1, 2},
        {2, 3},
        {3, 0},
        {4, 5},
        {5, 6},
        {6, 7},
        {7, 4},
        {0, 4},
        {1, 5},
        {2, 6},
        {3, 7}
    };

    // For any edge, if one vertex is inside the surface and the other is outside the surface then the edge intersects the surface
    // For each of the 8 vertices of the cube can be two possible states: either inside or outside the surface
    // For any cube, there are 2^8=256 possible sets of vertex states
    // This table lists the edges intersected by the surface for all 256 possible vertex states
    // There are 12 edges.  For each entry in the table, if edge #n is intersected, then bit #n is set to 1
    const GLint CubeEdgeFlags[256] = {
        0x000, 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
        0x190, 0x099, 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
        0x230, 0x339, 0x033, 0x13a, 0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
        0x3a0, 0x2a9, 0x1a3, 0x0aa, 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
        0x460, 0x569, 0x663, 0x76a, 0x066, 0x16f, 0x265, 0x36c, 0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
        0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0x0ff, 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
        0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x055, 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
        0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0x0cc, 0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
        0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc, 0x0cc, 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
        0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x055, 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
        0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc, 0x2fc, 0x3f5, 0x0ff, 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
        0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x066, 0x76a, 0x663, 0x569, 0x460,
        0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6, 0x0aa, 0x1a3, 0x2a9, 0x3a0,
        0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c, 0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x033, 0x339, 0x230,
        0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x099, 0x190,
        0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x000};

    // For each of the possible vertex states listed in CubeEdgeFlags there is a specific triangulation of the edge intersection points.
    // TriangleConnectionTable lists all of them in the form of 0-5 edge triples with the list terminated by the invalid value -1.
    // For example:
    // TriangleConnectionTable[3] list the 2 triangles formed when corner[0] and corner[1] are inside of the surface, but the rest of the cube is not .
    static const GLint TriangleConnectionTable[256][16] = {
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  1,  9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  8,  3,  9,  8,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  3,  1,  2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 9,  2, 10,  0,  2,  9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 2,  8,  3,  2, 10,  8, 10,  9,  8, -1, -1, -1, -1, -1, -1, -1},
        { 3, 11,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0, 11,  2,  8, 11,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  9,  0,  2,  3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1, 11,  2,  1,  9, 11,  9,  8, 11, -1, -1, -1, -1, -1, -1, -1},
        { 3, 10,  1, 11, 10,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0, 10,  1,  0,  8, 10,  8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        { 3,  9,  0,  3, 11,  9, 11, 10,  9, -1, -1, -1, -1, -1, -1, -1},
        { 9,  8, 10, 10,  8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  7,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  3,  0,  7,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  1,  9,  8,  4,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  1,  9,  4,  7,  1,  7,  3,  1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 10,  8,  4,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 3,  4,  7,  3,  0,  4,  1,  2, 10, -1, -1, -1, -1, -1, -1, -1},
        { 9,  2, 10,  9,  0,  2,  8,  4,  7, -1, -1, -1, -1, -1, -1, -1},
        { 2, 10,  9,  2,  9,  7,  2,  7,  3,  7,  9,  4, -1, -1, -1, -1},
        { 8,  4,  7,  3, 11,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11,  4,  7, 11,  2,  4,  2,  0,  4, -1, -1, -1, -1, -1, -1, -1},
        { 9,  0,  1,  8,  4,  7,  2,  3, 11, -1, -1, -1, -1, -1, -1, -1},
        { 4,  7, 11,  9,  4, 11,  9, 11,  2,  9,  2,  1, -1, -1, -1, -1},
        { 3, 10,  1,  3, 11, 10,  7,  8,  4, -1, -1, -1, -1, -1, -1, -1},
        { 1, 11, 10,  1,  4, 11,  1,  0,  4,  7, 11,  4, -1, -1, -1, -1},
        { 4,  7,  8,  9,  0, 11,  9, 11, 10, 11,  0,  3, -1, -1, -1, -1},
        { 4,  7, 11,  4, 11,  9,  9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        { 9,  5,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 9,  5,  4,  0,  8,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  5,  4,  1,  5,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 8,  5,  4,  8,  3,  5,  3,  1,  5, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 10,  9,  5,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 3,  0,  8,  1,  2, 10,  4,  9,  5, -1, -1, -1, -1, -1, -1, -1},
        { 5,  2, 10,  5,  4,  2,  4,  0,  2, -1, -1, -1, -1, -1, -1, -1},
        { 2, 10,  5,  3,  2,  5,  3,  5,  4,  3,  4,  8, -1, -1, -1, -1},
        { 9,  5,  4,  2,  3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0, 11,  2,  0,  8, 11,  4,  9,  5, -1, -1, -1, -1, -1, -1, -1},
        { 0,  5,  4,  0,  1,  5,  2,  3, 11, -1, -1, -1, -1, -1, -1, -1},
        { 2,  1,  5,  2,  5,  8,  2,  8, 11,  4,  8,  5, -1, -1, -1, -1},
        {10,  3, 11, 10,  1,  3,  9,  5,  4, -1, -1, -1, -1, -1, -1, -1},
        { 4,  9,  5,  0,  8,  1,  8, 10,  1,  8, 11, 10, -1, -1, -1, -1},
        { 5,  4,  0,  5,  0, 11,  5, 11, 10, 11,  0,  3, -1, -1, -1, -1},
        { 5,  4,  8,  5,  8, 10, 10,  8, 11, -1, -1, -1, -1, -1, -1, -1},
        { 9,  7,  8,  5,  7,  9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 9,  3,  0,  9,  5,  3,  5,  7,  3, -1, -1, -1, -1, -1, -1, -1},
        { 0,  7,  8,  0,  1,  7,  1,  5,  7, -1, -1, -1, -1, -1, -1, -1},
        { 1,  5,  3,  3,  5,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 9,  7,  8,  9,  5,  7, 10,  1,  2, -1, -1, -1, -1, -1, -1, -1},
        {10,  1,  2,  9,  5,  0,  5,  3,  0,  5,  7,  3, -1, -1, -1, -1},
        { 8,  0,  2,  8,  2,  5,  8,  5,  7, 10,  5,  2, -1, -1, -1, -1},
        { 2, 10,  5,  2,  5,  3,  3,  5,  7, -1, -1, -1, -1, -1, -1, -1},
        { 7,  9,  5,  7,  8,  9,  3, 11,  2, -1, -1, -1, -1, -1, -1, -1},
        { 9,  5,  7,  9,  7,  2,  9,  2,  0,  2,  7, 11, -1, -1, -1, -1},
        { 2,  3, 11,  0,  1,  8,  1,  7,  8,  1,  5,  7, -1, -1, -1, -1},
        {11,  2,  1, 11,  1,  7,  7,  1,  5, -1, -1, -1, -1, -1, -1, -1},
        { 9,  5,  8,  8,  5,  7, 10,  1,  3, 10,  3, 11, -1, -1, -1, -1},
        { 5,  7,  0,  5,  0,  9,  7, 11,  0,  1,  0, 10, 11, 10,  0, -1},
        {11, 10,  0, 11,  0,  3, 10,  5,  0,  8,  0,  7,  5,  7,  0, -1},
        {11, 10,  5,  7, 11,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10,  6,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  3,  5, 10,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 9,  0,  1,  5, 10,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  8,  3,  1,  9,  8,  5, 10,  6, -1, -1, -1, -1, -1, -1, -1},
        { 1,  6,  5,  2,  6,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  6,  5,  1,  2,  6,  3,  0,  8, -1, -1, -1, -1, -1, -1, -1},
        { 9,  6,  5,  9,  0,  6,  0,  2,  6, -1, -1, -1, -1, -1, -1, -1},
        { 5,  9,  8,  5,  8,  2,  5,  2,  6,  3,  2,  8, -1, -1, -1, -1},
        { 2,  3, 11, 10,  6,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11,  0,  8, 11,  2,  0, 10,  6,  5, -1, -1, -1, -1, -1, -1, -1},
        { 0,  1,  9,  2,  3, 11,  5, 10,  6, -1, -1, -1, -1, -1, -1, -1},
        { 5, 10,  6,  1,  9,  2,  9, 11,  2,  9,  8, 11, -1, -1, -1, -1},
        { 6,  3, 11,  6,  5,  3,  5,  1,  3, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8, 11,  0, 11,  5,  0,  5,  1,  5, 11,  6, -1, -1, -1, -1},
        { 3, 11,  6,  0,  3,  6,  0,  6,  5,  0,  5,  9, -1, -1, -1, -1},
        { 6,  5,  9,  6,  9, 11, 11,  9,  8, -1, -1, -1, -1, -1, -1, -1},
        { 5, 10,  6,  4,  7,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  3,  0,  4,  7,  3,  6,  5, 10, -1, -1, -1, -1, -1, -1, -1},
        { 1,  9,  0,  5, 10,  6,  8,  4,  7, -1, -1, -1, -1, -1, -1, -1},
        {10,  6,  5,  1,  9,  7,  1,  7,  3,  7,  9,  4, -1, -1, -1, -1},
        { 6,  1,  2,  6,  5,  1,  4,  7,  8, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2,  5,  5,  2,  6,  3,  0,  4,  3,  4,  7, -1, -1, -1, -1},
        { 8,  4,  7,  9,  0,  5,  0,  6,  5,  0,  2,  6, -1, -1, -1, -1},
        { 7,  3,  9,  7,  9,  4,  3,  2,  9,  5,  9,  6,  2,  6,  9, -1},
        { 3, 11,  2,  7,  8,  4, 10,  6,  5, -1, -1, -1, -1, -1, -1, -1},
        { 5, 10,  6,  4,  7,  2,  4,  2,  0,  2,  7, 11, -1, -1, -1, -1},
        { 0,  1,  9,  4,  7,  8,  2,  3, 11,  5, 10,  6, -1, -1, -1, -1},
        { 9,  2,  1,  9, 11,  2,  9,  4, 11,  7, 11,  4,  5, 10,  6, -1},
        { 8,  4,  7,  3, 11,  5,  3,  5,  1,  5, 11,  6, -1, -1, -1, -1},
        { 5,  1, 11,  5, 11,  6,  1,  0, 11,  7, 11,  4,  0,  4, 11, -1},
        { 0,  5,  9,  0,  6,  5,  0,  3,  6, 11,  6,  3,  8,  4,  7, -1},
        { 6,  5,  9,  6,  9, 11,  4,  7,  9,  7, 11,  9, -1, -1, -1, -1},
        {10,  4,  9,  6,  4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4, 10,  6,  4,  9, 10,  0,  8,  3, -1, -1, -1, -1, -1, -1, -1},
        {10,  0,  1, 10,  6,  0,  6,  4,  0, -1, -1, -1, -1, -1, -1, -1},
        { 8,  3,  1,  8,  1,  6,  8,  6,  4,  6,  1, 10, -1, -1, -1, -1},
        { 1,  4,  9,  1,  2,  4,  2,  6,  4, -1, -1, -1, -1, -1, -1, -1},
        { 3,  0,  8,  1,  2,  9,  2,  4,  9,  2,  6,  4, -1, -1, -1, -1},
        { 0,  2,  4,  4,  2,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 8,  3,  2,  8,  2,  4,  4,  2,  6, -1, -1, -1, -1, -1, -1, -1},
        {10,  4,  9, 10,  6,  4, 11,  2,  3, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  2,  2,  8, 11,  4,  9, 10,  4, 10,  6, -1, -1, -1, -1},
        { 3, 11,  2,  0,  1,  6,  0,  6,  4,  6,  1, 10, -1, -1, -1, -1},
        { 6,  4,  1,  6,  1, 10,  4,  8,  1,  2,  1, 11,  8, 11,  1, -1},
        { 9,  6,  4,  9,  3,  6,  9,  1,  3, 11,  6,  3, -1, -1, -1, -1},
        { 8, 11,  1,  8,  1,  0, 11,  6,  1,  9,  1,  4,  6,  4,  1, -1},
        { 3, 11,  6,  3,  6,  0,  0,  6,  4, -1, -1, -1, -1, -1, -1, -1},
        { 6,  4,  8, 11,  6,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 7, 10,  6,  7,  8, 10,  8,  9, 10, -1, -1, -1, -1, -1, -1, -1},
        { 0,  7,  3,  0, 10,  7,  0,  9, 10,  6,  7, 10, -1, -1, -1, -1},
        {10,  6,  7,  1, 10,  7,  1,  7,  8,  1,  8,  0, -1, -1, -1, -1},
        {10,  6,  7, 10,  7,  1,  1,  7,  3, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2,  6,  1,  6,  8,  1,  8,  9,  8,  6,  7, -1, -1, -1, -1},
        { 2,  6,  9,  2,  9,  1,  6,  7,  9,  0,  9,  3,  7,  3,  9, -1},
        { 7,  8,  0,  7,  0,  6,  6,  0,  2, -1, -1, -1, -1, -1, -1, -1},
        { 7,  3,  2,  6,  7,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 2,  3, 11, 10,  6,  8, 10,  8,  9,  8,  6,  7, -1, -1, -1, -1},
        { 2,  0,  7,  2,  7, 11,  0,  9,  7,  6,  7, 10,  9, 10,  7, -1},
        { 1,  8,  0,  1,  7,  8,  1, 10,  7,  6,  7, 10,  2,  3, 11, -1},
        {11,  2,  1, 11,  1,  7, 10,  6,  1,  6,  7,  1, -1, -1, -1, -1},
        { 8,  9,  6,  8,  6,  7,  9,  1,  6, 11,  6,  3,  1,  3,  6, -1},
        { 0,  9,  1, 11,  6,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 7,  8,  0,  7,  0,  6,  3, 11,  0, 11,  6,  0, -1, -1, -1, -1},
        { 7, 11,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 7,  6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 3,  0,  8, 11,  7,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  1,  9, 11,  7,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 8,  1,  9,  8,  3,  1, 11,  7,  6, -1, -1, -1, -1, -1, -1, -1},
        {10,  1,  2,  6, 11,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 10,  3,  0,  8,  6, 11,  7, -1, -1, -1, -1, -1, -1, -1},
        { 2,  9,  0,  2, 10,  9,  6, 11,  7, -1, -1, -1, -1, -1, -1, -1},
        { 6, 11,  7,  2, 10,  3, 10,  8,  3, 10,  9,  8, -1, -1, -1, -1},
        { 7,  2,  3,  6,  2,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 7,  0,  8,  7,  6,  0,  6,  2,  0, -1, -1, -1, -1, -1, -1, -1},
        { 2,  7,  6,  2,  3,  7,  0,  1,  9, -1, -1, -1, -1, -1, -1, -1},
        { 1,  6,  2,  1,  8,  6,  1,  9,  8,  8,  7,  6, -1, -1, -1, -1},
        {10,  7,  6, 10,  1,  7,  1,  3,  7, -1, -1, -1, -1, -1, -1, -1},
        {10,  7,  6,  1,  7, 10,  1,  8,  7,  1,  0,  8, -1, -1, -1, -1},
        { 0,  3,  7,  0,  7, 10,  0, 10,  9,  6, 10,  7, -1, -1, -1, -1},
        { 7,  6, 10,  7, 10,  8,  8, 10,  9, -1, -1, -1, -1, -1, -1, -1},
        { 6,  8,  4, 11,  8,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 3,  6, 11,  3,  0,  6,  0,  4,  6, -1, -1, -1, -1, -1, -1, -1},
        { 8,  6, 11,  8,  4,  6,  9,  0,  1, -1, -1, -1, -1, -1, -1, -1},
        { 9,  4,  6,  9,  6,  3,  9,  3,  1, 11,  3,  6, -1, -1, -1, -1},
        { 6,  8,  4,  6, 11,  8,  2, 10,  1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 10,  3,  0, 11,  0,  6, 11,  0,  4,  6, -1, -1, -1, -1},
        { 4, 11,  8,  4,  6, 11,  0,  2,  9,  2, 10,  9, -1, -1, -1, -1},
        {10,  9,  3, 10,  3,  2,  9,  4,  3, 11,  3,  6,  4,  6,  3, -1},
        { 8,  2,  3,  8,  4,  2,  4,  6,  2, -1, -1, -1, -1, -1, -1, -1},
        { 0,  4,  2,  4,  6,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  9,  0,  2,  3,  4,  2,  4,  6,  4,  3,  8, -1, -1, -1, -1},
        { 1,  9,  4,  1,  4,  2,  2,  4,  6, -1, -1, -1, -1, -1, -1, -1},
        { 8,  1,  3,  8,  6,  1,  8,  4,  6,  6, 10,  1, -1, -1, -1, -1},
        {10,  1,  0, 10,  0,  6,  6,  0,  4, -1, -1, -1, -1, -1, -1, -1},
        { 4,  6,  3,  4,  3,  8,  6, 10,  3,  0,  3,  9, 10,  9,  3, -1},
        {10,  9,  4,  6, 10,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  9,  5,  7,  6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  3,  4,  9,  5, 11,  7,  6, -1, -1, -1, -1, -1, -1, -1},
        { 5,  0,  1,  5,  4,  0,  7,  6, 11, -1, -1, -1, -1, -1, -1, -1},
        {11,  7,  6,  8,  3,  4,  3,  5,  4,  3,  1,  5, -1, -1, -1, -1},
        { 9,  5,  4, 10,  1,  2,  7,  6, 11, -1, -1, -1, -1, -1, -1, -1},
        { 6, 11,  7,  1,  2, 10,  0,  8,  3,  4,  9,  5, -1, -1, -1, -1},
        { 7,  6, 11,  5,  4, 10,  4,  2, 10,  4,  0,  2, -1, -1, -1, -1},
        { 3,  4,  8,  3,  5,  4,  3,  2,  5, 10,  5,  2, 11,  7,  6, -1},
        { 7,  2,  3,  7,  6,  2,  5,  4,  9, -1, -1, -1, -1, -1, -1, -1},
        { 9,  5,  4,  0,  8,  6,  0,  6,  2,  6,  8,  7, -1, -1, -1, -1},
        { 3,  6,  2,  3,  7,  6,  1,  5,  0,  5,  4,  0, -1, -1, -1, -1},
        { 6,  2,  8,  6,  8,  7,  2,  1,  8,  4,  8,  5,  1,  5,  8, -1},
        { 9,  5,  4, 10,  1,  6,  1,  7,  6,  1,  3,  7, -1, -1, -1, -1},
        { 1,  6, 10,  1,  7,  6,  1,  0,  7,  8,  7,  0,  9,  5,  4, -1},
        { 4,  0, 10,  4, 10,  5,  0,  3, 10,  6, 10,  7,  3,  7, 10, -1},
        { 7,  6, 10,  7, 10,  8,  5,  4, 10,  4,  8, 10, -1, -1, -1, -1},
        { 6,  9,  5,  6, 11,  9, 11,  8,  9, -1, -1, -1, -1, -1, -1, -1},
        { 3,  6, 11,  0,  6,  3,  0,  5,  6,  0,  9,  5, -1, -1, -1, -1},
        { 0, 11,  8,  0,  5, 11,  0,  1,  5,  5,  6, 11, -1, -1, -1, -1},
        { 6, 11,  3,  6,  3,  5,  5,  3,  1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 10,  9,  5, 11,  9, 11,  8, 11,  5,  6, -1, -1, -1, -1},
        { 0, 11,  3,  0,  6, 11,  0,  9,  6,  5,  6,  9,  1,  2, 10, -1},
        {11,  8,  5, 11,  5,  6,  8,  0,  5, 10,  5,  2,  0,  2,  5, -1},
        { 6, 11,  3,  6,  3,  5,  2, 10,  3, 10,  5,  3, -1, -1, -1, -1},
        { 5,  8,  9,  5,  2,  8,  5,  6,  2,  3,  8,  2, -1, -1, -1, -1},
        { 9,  5,  6,  9,  6,  0,  0,  6,  2, -1, -1, -1, -1, -1, -1, -1},
        { 1,  5,  8,  1,  8,  0,  5,  6,  8,  3,  8,  2,  6,  2,  8, -1},
        { 1,  5,  6,  2,  1,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  3,  6,  1,  6, 10,  3,  8,  6,  5,  6,  9,  8,  9,  6, -1},
        {10,  1,  0, 10,  0,  6,  9,  5,  0,  5,  6,  0, -1, -1, -1, -1},
        { 0,  3,  8,  5,  6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10,  5,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11,  5, 10,  7,  5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11,  5, 10, 11,  7,  5,  8,  3,  0, -1, -1, -1, -1, -1, -1, -1},
        { 5, 11,  7,  5, 10, 11,  1,  9,  0, -1, -1, -1, -1, -1, -1, -1},
        {10,  7,  5, 10, 11,  7,  9,  8,  1,  8,  3,  1, -1, -1, -1, -1},
        {11,  1,  2, 11,  7,  1,  7,  5,  1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  3,  1,  2,  7,  1,  7,  5,  7,  2, 11, -1, -1, -1, -1},
        { 9,  7,  5,  9,  2,  7,  9,  0,  2,  2, 11,  7, -1, -1, -1, -1},
        { 7,  5,  2,  7,  2, 11,  5,  9,  2,  3,  2,  8,  9,  8,  2, -1},
        { 2,  5, 10,  2,  3,  5,  3,  7,  5, -1, -1, -1, -1, -1, -1, -1},
        { 8,  2,  0,  8,  5,  2,  8,  7,  5, 10,  2,  5, -1, -1, -1, -1},
        { 9,  0,  1,  5, 10,  3,  5,  3,  7,  3, 10,  2, -1, -1, -1, -1},
        { 9,  8,  2,  9,  2,  1,  8,  7,  2, 10,  2,  5,  7,  5,  2, -1},
        { 1,  3,  5,  3,  7,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  7,  0,  7,  1,  1,  7,  5, -1, -1, -1, -1, -1, -1, -1},
        { 9,  0,  3,  9,  3,  5,  5,  3,  7, -1, -1, -1, -1, -1, -1, -1},
        { 9,  8,  7,  5,  9,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 5,  8,  4,  5, 10,  8, 10, 11,  8, -1, -1, -1, -1, -1, -1, -1},
        { 5,  0,  4,  5, 11,  0,  5, 10, 11, 11,  3,  0, -1, -1, -1, -1},
        { 0,  1,  9,  8,  4, 10,  8, 10, 11, 10,  4,  5, -1, -1, -1, -1},
        {10, 11,  4, 10,  4,  5, 11,  3,  4,  9,  4,  1,  3,  1,  4, -1},
        { 2,  5,  1,  2,  8,  5,  2, 11,  8,  4,  5,  8, -1, -1, -1, -1},
        { 0,  4, 11,  0, 11,  3,  4,  5, 11,  2, 11,  1,  5,  1, 11, -1},
        { 0,  2,  5,  0,  5,  9,  2, 11,  5,  4,  5,  8, 11,  8,  5, -1},
        { 9,  4,  5,  2, 11,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 2,  5, 10,  3,  5,  2,  3,  4,  5,  3,  8,  4, -1, -1, -1, -1},
        { 5, 10,  2,  5,  2,  4,  4,  2,  0, -1, -1, -1, -1, -1, -1, -1},
        { 3, 10,  2,  3,  5, 10,  3,  8,  5,  4,  5,  8,  0,  1,  9, -1},
        { 5, 10,  2,  5,  2,  4,  1,  9,  2,  9,  4,  2, -1, -1, -1, -1},
        { 8,  4,  5,  8,  5,  3,  3,  5,  1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  4,  5,  1,  0,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 8,  4,  5,  8,  5,  3,  9,  0,  5,  0,  3,  5, -1, -1, -1, -1},
        { 9,  4,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4, 11,  7,  4,  9, 11,  9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        { 0,  8,  3,  4,  9,  7,  9, 11,  7,  9, 10, 11, -1, -1, -1, -1},
        { 1, 10, 11,  1, 11,  4,  1,  4,  0,  7,  4, 11, -1, -1, -1, -1},
        { 3,  1,  4,  3,  4,  8,  1, 10,  4,  7,  4, 11, 10, 11,  4, -1},
        { 4, 11,  7,  9, 11,  4,  9,  2, 11,  9,  1,  2, -1, -1, -1, -1},
        { 9,  7,  4,  9, 11,  7,  9,  1, 11,  2, 11,  1,  0,  8,  3, -1},
        {11,  7,  4, 11,  4,  2,  2,  4,  0, -1, -1, -1, -1, -1, -1, -1},
        {11,  7,  4, 11,  4,  2,  8,  3,  4,  3,  2,  4, -1, -1, -1, -1},
        { 2,  9, 10,  2,  7,  9,  2,  3,  7,  7,  4,  9, -1, -1, -1, -1},
        { 9, 10,  7,  9,  7,  4, 10,  2,  7,  8,  7,  0,  2,  0,  7, -1},
        { 3,  7, 10,  3, 10,  2,  7,  4, 10,  1, 10,  0,  4,  0, 10, -1},
        { 1, 10,  2,  8,  7,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  9,  1,  4,  1,  7,  7,  1,  3, -1, -1, -1, -1, -1, -1, -1},
        { 4,  9,  1,  4,  1,  7,  0,  8,  1,  8,  7,  1, -1, -1, -1, -1},
        { 4,  0,  3,  7,  4,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 4,  8,  7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 9, 10,  8, 10, 11,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 3,  0,  9,  3,  9, 11, 11,  9, 10, -1, -1, -1, -1, -1, -1, -1},
        { 0,  1, 10,  0, 10,  8,  8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        { 3,  1, 10, 11,  3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  2, 11,  1, 11,  9,  9, 11,  8, -1, -1, -1, -1, -1, -1, -1},
        { 3,  0,  9,  3,  9, 11,  1,  2,  9,  2, 11,  9, -1, -1, -1, -1},
        { 0,  2, 11,  8,  0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 3,  2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 2,  3,  8,  2,  8, 10, 10,  8,  9, -1, -1, -1, -1, -1, -1, -1},
        { 9, 10,  2,  0,  9,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 2,  3,  8,  2,  8, 10,  0,  1,  8,  1, 10,  8, -1, -1, -1, -1},
        { 1, 10,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 1,  3,  8,  9,  1,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  9,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        { 0,  3,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
    };
}

// ======================================================================
// Help and Log Functions
// ======================================================================

GLvoid MarchingCubesGL::PrintHelp()
{
    std::cout << "\t1                        show internal grid\n";
    std::cout << "\t2                        show external grid\n";
    std::cout << "\t3                        show internal points in red\n";
    std::cout << "\t4                        show internal points in white\n";
    std::cout << "\t5                        show external points in red\n";
    std::cout << "\t6                        show external points in white\n";

    std::cout << "\tPageUp/PageDown          increase/decrease surface value\n";
    std::cout << "\tShift + PageUp/PageDown  slowly increase/decrease surface value\n";

    std::cout << "\ts                        change sample function forward\n";
    std::cout << "\tShift + s                change sample function backward\n";

    std::cout << "\tw                        wireframe on/off\n";
    std::cout << "\tl                        toggle lighting / normal colors\n";

    std::cout << "\tHome                     spin scene on/off\n";
    std::cout << "\tShift + Home             slowly spin scene on/off\n";
    std::cout << "\tEnd                      source point animation on/off\n";
    std::cout << "\tShift + End              slowly source point animation on/off\n";
}

GLvoid MarchingCubesGL::PrintLog()
{
    std::clog << "\r                                                                      ";
    std::clog << "\rSurfaceValue: " << SurfaceValue
              << " RowSize: " << RowSize
              << " StepSize: " << StepSize
              << std::flush;
}

// ======================================================================
// Scalar Field Functions
// ======================================================================

// Sphere in the center of the volume.
GLfloat MarchingCubesGL::Sample1(const GLfloat x, const GLfloat y, const GLfloat z)
{
    GLdouble          result = 0.0;
    constexpr GLfloat scale  = 10.0f;
    GLdouble          dx, dy, dz;
    dx = x - 0.5;
    dy = y - 0.5;
    dz = z - 0.5;
    result += (dx * dx + dy * dy + dz * dz) * scale;

    return static_cast<GLfloat>(result);
}

// Distance of (x, y, z) from three moving points. (Spheres)
GLfloat MarchingCubesGL::Sample2(const GLfloat x, const GLfloat y, const GLfloat z)
{
    GLdouble          result = 0.0;
    constexpr GLfloat scale  = 40.0f;
    GLdouble          dx, dy, dz;
    dx = x - SourcePoint[0].X;
    dy = y - SourcePoint[0].Y;
    dz = z - SourcePoint[0].Z;
    result += 0.5 / (dx * dx + dy * dy + dz * dz);

    dx = x - SourcePoint[1].X;
    dy = y - SourcePoint[1].Y;
    dz = z - SourcePoint[1].Z;
    result += 1.0 / (dx * dx + dy * dy + dz * dz);

    dx = x - SourcePoint[2].X;
    dy = y - SourcePoint[2].Y;
    dz = z - SourcePoint[2].Z;
    result += 1.5 / (dx * dx + dy * dy + dz * dz);

    return static_cast<GLfloat>(1.0f / result) * scale;
}

// Distance of (x, y, z) from three moving points. (Cylinders)
GLfloat MarchingCubesGL::Sample3(const GLfloat x, const GLfloat y, const GLfloat z)
{
    GLdouble          result = 0.0;
    constexpr GLfloat scale  = 30.0f;
    GLdouble          dx, dy, dz;
    dx = x - SourcePoint[0].X;
    dy = y - SourcePoint[0].Y;
    result += 0.5 / (dx * dx + dy * dy);

    dx = x - SourcePoint[1].X;
    dz = z - SourcePoint[1].Z;
    result += 0.75 / (dx * dx + dz * dz);

    dy = y - SourcePoint[2].Y;
    dz = z - SourcePoint[2].Z;
    result += 1.0 / (dy * dy + dz * dz);

    return static_cast<GLfloat>(3.0f / result) * scale;
}

// Defines a height field by plugging the distance from the center into the sin and cos functions.
GLfloat MarchingCubesGL::Sample4(const GLfloat x, const GLfloat y, const GLfloat z)
{
    GLfloat height;
    height = 20.0f * (Time + sqrtf((0.5f - x) * (0.5f - x) + (0.5f - y) * (0.5f - y)));
    height = 1.5f + 0.1f * (sinf(height) + cosf(height));

    return height - z;
}

// Distance of (x, y, z) from two moving points.
GLfloat MarchingCubesGL::Sample5(const GLfloat x, const GLfloat y, const GLfloat z)
{
    GLdouble          result = 0.0;
    constexpr GLfloat scale  = 50.0f;
    GLdouble          dx, dy, dz;
    dx = x - SourcePoint[0].X;
    dy = y - SourcePoint[0].Y;
    dz = z - SourcePoint[0].Z;
    result += 2.0 / (dx * dx + dy * dy + dz * dz);

    dx = x - SourcePoint[1].X;
    dy = y - SourcePoint[1].Y;
    dz = z - SourcePoint[1].Z;
    result += 2.0 / (dx * dx + dy * dy + dz * dz);
    return static_cast<GLfloat>(1.0f / result) * scale;
}

// Bobby Molecules
GLfloat MarchingCubesGL::Sample6(const GLfloat x, const GLfloat y, const GLfloat z)
{
    GLdouble          result = 0.0;
    constexpr GLfloat scale  = 30.0f;
    GLdouble          dx, dy, dz;
    dx = x - 0.7;
    dy = y - 0.7;
    dz = z - 0.5;
    result += 1.0 / (dx * dx + dy * dy + dz * dz);

    dx = x - 0.3;
    dy = y - 0.3;
    dz = z - 0.5;
    result += 2.0 / (dx * dx + dy * dy + dz * dz);

    return static_cast<GLfloat>(1.0f / result) * scale;
}

// Potential Surface
GLfloat MarchingCubesGL::Sample7(const GLfloat x, const GLfloat y, const GLfloat z)
{
    const Dodecahedron d;
    GLdouble           result = 0.0;
    constexpr GLfloat  scale  = 2.5f;
    const GLfloat      dx     = 0.5f - x;
    const GLfloat      dy     = 0.5f - y;
    const GLfloat      dz     = 0.5f - z;

    // Each Dodecahedron contributes to Potential
    for(const auto& v : d.Data)
    {
        result += 1.0f / sqrt((dx + v[0]) * (dx + v[0]) + (dy + v[1]) * (dy + v[1]) + (dz + v[2]) * (dz + v[2]));
    }
    return static_cast<GLfloat>(1.0f / (0.0397f * result)) * scale;
}

// Toroidal Surface
GLfloat MarchingCubesGL::Sample8(const GLfloat x, const GLfloat y, const GLfloat z)
{
    constexpr GLfloat a     = 2.5f;
    constexpr GLfloat scale = 7.0f;
    const GLfloat     dx    = scale * (0.5f - x);
    const GLfloat     dy    = scale * (0.5f - y);
    const GLfloat     dz    = scale * (0.5f - z);

    const GLdouble result = (pow(sqrt(pow(dx, 2) + pow(dy, 2)) - a, 2) + pow(dz, 2))
                          * (pow(sqrt(pow(dy, 2) + pow(dz, 2)) - a, 2) + pow(dx, 2))
                          * (pow(sqrt(pow(dz, 2) + pow(dx, 2)) - a, 2) + pow(dy, 2));
    return static_cast<GLfloat>(result);
}

// Double Torus
GLfloat MarchingCubesGL::Sample9(const GLfloat x, const GLfloat y, const GLfloat z)
{
    constexpr GLfloat scale = 2.25f;
    const GLfloat     dx    = scale * (0.5f - x);
    const GLfloat     dy    = scale * (0.5f - y);
    const GLfloat     dz    = scale * (0.5f - z);

    const GLdouble result = pow(dx, 8)
                          + pow(dx, 4)
                          - 2 * pow(dx, 6)
                          - 2 * pow(dx, 2) * pow(dy, 2)
                          + 2 * pow(dx, 4) * pow(dy, 2)
                          + pow(dy, 4)
                          + pow(dz, 2);

    return static_cast<GLfloat>(result) * 30.0f;
}

// Chmutov Surface. N = 4
GLfloat MarchingCubesGL::Sample10(const GLfloat x, const GLfloat y, const GLfloat z)
{
    constexpr GLfloat scale = 2.5f;
    const GLfloat     dx    = scale * (0.5f - x);
    const GLfloat     dy    = scale * (0.5f - y);
    const GLfloat     dz    = scale * (0.5f - z);

    constexpr GLdouble factor = 3.0;
    const GLdouble     result = factor * ((pow(dx, 2) + pow(dy, 2) + pow(dz, 2)) - (pow(dx, 4) + pow(dy, 4) + pow(dz, 4)));

    return static_cast<GLfloat>(result);    // There is some problem with this function.
}

// Chmutov Surface. N = 6
GLfloat MarchingCubesGL::Sample11(const GLfloat x, const GLfloat y, const GLfloat z)
{
    constexpr GLfloat scale = 2.5f;
    const GLfloat     dx    = scale * (0.5f - x);
    const GLfloat     dy    = scale * (0.5f - y);
    const GLfloat     dz    = scale * (0.5f - z);

    constexpr GLdouble factor = 0.7;
    const GLdouble     result = factor * (pow(dx, 2) * pow(3 - 4 * pow(dx, 2), 2) + pow(dy, 2) * pow(3 - 4 * pow(dy, 2), 2) + pow(dz, 2) * pow(3 - 4 * pow(dz, 2), 2));

    return static_cast<GLfloat>(result);
}

// ======================================================================
// Marching Cubes
// ======================================================================

// MarchCube performs the Marching Cubes algorithm on a single cube
GLvoid MarchingCubesGL::MarchCube(const GLfloat x, const GLfloat y, const GLfloat z, const GLfloat stepSize)
{
    GLfloat  cube[8];         // The surface value at each of the 8 corners of the cube
    GLvector edgeVertex[12];  // The position of the vertex on the edge
    GLvector edgeNorm[12];    // The normal of the vertex on the edge
    GLint    flagIndex = 0;   // The edge flags. The bit-mask is: BBBB BBBB BBBB (dec: 0-4095).
                              // If a bit is set, the vertex at the edge is inside of the surface.
    for(GLint vertex = 0; vertex < 8; vertex++)
    {
        // Make a local copy of the values at the cube's corners
        cube[vertex] = Sample(x + VertexOffset[vertex][0] * stepSize,
                              y + VertexOffset[vertex][1] * stepSize,
                              z + VertexOffset[vertex][2] * stepSize);

        // Find which vertices are inside of the surface and which are outside
        if(cube[vertex] <= SurfaceValue)
        {
            flagIndex |= 1 << vertex;
            if(PointsInternal)
            {
                DrawPoint(x + VertexOffset[vertex][0] * stepSize,
                          y + VertexOffset[vertex][1] * stepSize,
                          z + VertexOffset[vertex][2] * stepSize,
                          PointsInternalWhite);
            }

            if(BBoxInternal)
            {
                DrawBoundingBox(x + VertexOffset[0][0] * stepSize,
                                y + VertexOffset[0][1] * stepSize,
                                z + VertexOffset[0][2] * stepSize,
                                stepSize);
            }
        }
        else
        {
            flagIndex |= 0 << vertex;

            if(PointsExtenal)
            {
                DrawPoint(x + VertexOffset[vertex][0] * stepSize,
                          y + VertexOffset[vertex][1] * stepSize,
                          z + VertexOffset[vertex][2] * stepSize,
                          PointsExtenalWhite);
            }

            if(BBoxExtenal)
            {
                DrawBoundingBox(x + VertexOffset[0][0] * stepSize,
                                y + VertexOffset[0][1] * stepSize,
                                z + VertexOffset[0][2] * stepSize,
                                stepSize);
            }
        }
    }

    // Find which edges are intersected by the surface
    const GLint edgeFlags = CubeEdgeFlags[flagIndex];

    // If the cube is entirely inside or outside of the surface, then there will be no intersections
    if(edgeFlags == 0)
    {
        return;
    }

    // Find the point of intersection of the surface with each edge
    // Then find the normal to the surface at those points
    for(GLint edge = 0; edge < 12; edge++)
    {
        // if there is an intersection on this edge: Vertex Interpolation
        if(edgeFlags & (1 << edge))
        {
            const GLvector p1 = {x + VertexOffset[EdgeConnection[edge][0]][0] * stepSize,
                                 y + VertexOffset[EdgeConnection[edge][0]][1] * stepSize,
                                 z + VertexOffset[EdgeConnection[edge][0]][2] * stepSize};

            const GLvector p2 = {x + VertexOffset[EdgeConnection[edge][1]][0] * stepSize,
                                 y + VertexOffset[EdgeConnection[edge][1]][1] * stepSize,
                                 z + VertexOffset[EdgeConnection[edge][1]][2] * stepSize};

            edgeVertex[edge] = VertexInterpolation(p1, p2, cube[EdgeConnection[edge][0]], cube[EdgeConnection[edge][1]], SurfaceValue);

            GetNormal(edgeNorm[edge], edgeVertex[edge].X, edgeVertex[edge].Y, edgeVertex[edge].Z);
        }
    }

    if(ShowSurface)
    {
        // Draw the triangles that were found. There can be up to five per cube.
        glBegin(GL_TRIANGLES);
        for(GLint i = 0; TriangleConnectionTable[flagIndex][i] != -1; i += 3)
        {
            // Vertices
            const GLint triangle1 = TriangleConnectionTable[flagIndex][i];
            const GLint triangle2 = TriangleConnectionTable[flagIndex][i + 1];
            const GLint triangle3 = TriangleConnectionTable[flagIndex][i + 2];

            // Colours
            GLvector color1;
            GetColor(color1, edgeVertex[triangle1], edgeNorm[triangle1]);
            GLvector color2;
            GetColor(color2, edgeVertex[triangle2], edgeNorm[triangle2]);
            GLvector color3;
            GetColor(color3, edgeVertex[triangle3], edgeNorm[triangle3]);

            // Draw Triangle: counter-clockwise
            glColor3f(color3.X, color3.Y, color3.Z);
            glNormal3f(edgeNorm[triangle3].X, edgeNorm[triangle3].Y, edgeNorm[triangle3].Z);
            glVertex3f(edgeVertex[triangle3].X, edgeVertex[triangle3].Y, edgeVertex[triangle3].Z);

            glColor3f(color2.X, color2.Y, color2.Z);
            glNormal3f(edgeNorm[triangle2].X, edgeNorm[triangle2].Y, edgeNorm[triangle2].Z);
            glVertex3f(edgeVertex[triangle2].X, edgeVertex[triangle2].Y, edgeVertex[triangle2].Z);

            glColor3f(color1.X, color1.Y, color1.Z);
            glNormal3f(edgeNorm[triangle1].X, edgeNorm[triangle1].Y, edgeNorm[triangle1].Z);
            glVertex3f(edgeVertex[triangle1].X, edgeVertex[triangle1].Y, edgeVertex[triangle1].Z);
        }
        glEnd();
    }
}

// MarchingCubes iterates over the entire dataset, calling MarchCube on each cube
GLvoid MarchingCubesGL::MarchingCubes()
{
    GLint x, y, z;
    for(x = 0; x < RowSize; x++)
    {
        for(y = 0; y < RowSize; y++)
        {
            for(z = 0; z < RowSize; z++)
            {
                MarchCube(
                    static_cast<GLfloat>(x) * StepSize,
                    static_cast<GLfloat>(y) * StepSize,
                    static_cast<GLfloat>(z) * StepSize,
                    StepSize);
            }
        }
    }
}

// ======================================================================
// Support Functions
// ======================================================================
// GetOffset finds the approximate point of intersection of the surface between two points with the values value1 and value2
GLfloat MarchingCubesGL::GetOffset(const GLfloat value1, const GLfloat value2, const GLfloat surfaceValue)
{
    const GLdouble delta = value2 - value1;

    if(delta == 0.0f)
    {
        return 0.5;
    }
    return (surfaceValue - value1) / static_cast<GLfloat>(delta);
}

// GetColor generates a color from a given position and normal of a point
GLvoid MarchingCubesGL::GetColor(GLvector& color, GLvector& position, const GLvector& normal)
{
    const GLfloat x = normal.X;
    const GLfloat y = normal.Y;
    const GLfloat z = normal.Z;
    color.X         = (x > 0.0f ? x : 0.0f) + (y < 0.0f ? -0.5f * y : 0.0f) + (z < 0.0f ? -0.5f * z : 0.0f);
    color.Y         = (y > 0.0f ? y : 0.0f) + (z < 0.0f ? -0.5f * z : 0.0f) + (x < 0.0f ? -0.5f * x : 0.0f);
    color.Z         = (z > 0.0f ? z : 0.0f) + (x < 0.0f ? -0.5f * x : 0.0f) + (y < 0.0f ? -0.5f * y : 0.0f);
}

// GetNormal() finds the gradient of the scalar field at a point
// This gradient can be used as a very accurate vertx normal for lighting calculations
GLvoid MarchingCubesGL::GetNormal(GLvector& normal, const GLfloat x, const GLfloat y, const GLfloat z)
{
    normal.X = Sample(x + 0.001f, y, z) - Sample(x - 0.001f, y, z);
    normal.Y = Sample(x, y + 0.001f, z) - Sample(x, y - 0.001f, z);
    normal.Z = Sample(x, y, z + 0.001f) - Sample(x, y, z - 0.001f);
    NormalizeVector(normal, normal);
}

GLvoid MarchingCubesGL::NormalizeVector(GLvector& destination, const GLvector& source)
{
    GLfloat length;
    GLfloat scale;

    length = sqrtf((source.X * source.X)
                   + (source.Y * source.Y)
                   + (source.Z * source.Z));

    if(length == 0.0)
    {
        destination.X = source.X;
        destination.Y = source.Y;
        destination.Z = source.Z;
    }
    else
    {
        scale         = 1.0f / length;
        destination.X = source.X * scale;
        destination.Y = source.Y * scale;
        destination.Z = source.Z * scale;
    }
}

// Sample data sets define scalal fields whose values vary by the X,Y and Z coordinate, and by the time value set by SetTime()
GLvoid MarchingCubesGL::SetTime(const GLfloat time)
{
    GLfloat offset;
    GLint   sourceNum;

    for(sourceNum = 0; sourceNum < 3; sourceNum++)
    {
        SourcePoint[sourceNum].X = 0.5;
        SourcePoint[sourceNum].Y = 0.5;
        SourcePoint[sourceNum].Z = 0.5;
    }

    Time   = time;
    offset = 1.0f + sinf(Time);
    SourcePoint[0].X *= offset;
    SourcePoint[1].Y *= offset;
    SourcePoint[2].Z *= offset;
}

// Return the point between two points in the same ratio as surfaceValue is between value1 and value2
MarchingCubesGL::GLvector MarchingCubesGL::VertexInterpolation(const GLvector p1, const GLvector p2, const GLfloat value1, const GLfloat value2, const GLfloat surfaceValue)
{
    GLvector p;

    if(std::abs(surfaceValue - value1) < 0.000001)
    {
        return p1;
    }

    if(std::abs(surfaceValue - value2) < 0.000001)
    {
        return p2;
    }

    if(std::abs(value1 - value2) < 0.000001)
    {
        return p1;
    }

    const GLdouble mu = (surfaceValue - value1) / (value2 - value1);

    p.X = p1.X + static_cast<GLfloat>(mu) * (p2.X - p1.X);
    p.Y = p1.Y + static_cast<GLfloat>(mu) * (p2.Y - p1.Y);
    p.Z = p1.Z + static_cast<GLfloat>(mu) * (p2.Z - p1.Z);

    return (p);
}

// ======================================================================
// OpenGL Functions
// ======================================================================
void MarchingCubesGL::Idle()
{
    glutPostRedisplay();
}

void MarchingCubesGL::Init(int argc, char** argv)
{
    constexpr GLsizei width  = 800;
    constexpr GLsizei height = 800;

    glutInit(&argc, argv);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutCreateWindow("Marching Cubes: Sample1");
    glutDisplayFunc(DrawScene);
    glutIdleFunc(Idle);
    glutReshapeFunc(Resize);
    glutKeyboardFunc(Keyboard);
    glutSpecialFunc(Special);

    glClearColor(0.4392f, 0.502f, 0.5647f, 1.0);    // Slate Gray
    glClearDepth(1.0);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT_AND_BACK, PolygonMode);

    glLightfv(GL_LIGHT0, GL_AMBIENT, LightColorAmbient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, LightColorDiffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, LightColorSpecular);
    glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0);

    glEnable(GL_LIGHT0);

    glMaterialfv(GL_FRONT, GL_AMBIENT, ColorFrontDiffuse);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, ColorFrontAmbient);
    glMaterialfv(GL_FRONT, GL_SPECULAR, ColorFrontSpecular);
    glMaterialf(GL_FRONT, GL_SHININESS, ColorFrontShininess);

    glMaterialfv(GL_BACK, GL_AMBIENT, ColorBackAmbient);
    glMaterialfv(GL_BACK, GL_DIFFUSE, ColorBackDiffuse);

    Resize(width, height);

    PrintHelp();
    glutMainLoop();
}

void MarchingCubesGL::DrawScene()
{
    static GLfloat pitch = 0.0;
    static GLfloat yaw   = 0.0;
    static GLfloat time  = 0.0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glPushMatrix();

    if(Spin)
    {
        pitch += SlowSpin ? 1.0f : 2.0f;
        yaw += SlowSpin ? 0.75f : 1.5f;
    }

    if(Move)
    {
        time += SlowMove ? 0.01f : 0.025f;
    }

    SetTime(time);

    glTranslatef(0.0, 0.0, -1.0);
    glRotatef(-pitch, 1.0, 0.0, 0.0);
    glRotatef(0.0, 0.0, 1.0, 0.0);
    glRotatef(yaw, 0.0, 0.0, 1.0);

    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_LIGHTING);
    glColor3f(0.9412f, 0.902f, 0.549f);    // Khaki
    glLineWidth(5.0f);
    glutWireCube(1.0);
    glLineWidth(1.0f);
    glPopAttrib();

    glPushMatrix();
    glTranslatef(-0.5, -0.5, -0.5);
    MarchingCubes();
    glPopMatrix();

    glPopMatrix();

    glutSwapBuffers();
}

void MarchingCubesGL::DrawPoint(const GLfloat x, const GLfloat y, const GLfloat z, const bool white)
{
    if(Light)
    {
        glDisable(GL_LIGHTING);
    }
    glPointSize(5);
    if(white)
    {
        glColor3f(1.0f, 1.0f, 1.0f);
    }
    else
    {
        glColor3f(1.0f, 0.3882f, 0.2784f);
    }

    glBegin(GL_POINTS);
    glVertex3f(x, y, z);
    glEnd();
    if(Light)
    {
        glEnable(GL_LIGHTING);
    }
}

GLvoid MarchingCubesGL::DrawBoundingBox(const GLfloat x, const GLfloat y, const GLfloat z, const GLfloat scale)
{
    if(Light)
    {
        glDisable(GL_LIGHTING);
    }
    glColor3f(1.0f, 1.0f, 1.0f);
    glLineWidth(1.0f);

    glBegin(GL_LINE_STRIP);
    glVertex3f(x, y, z);
    glVertex3f(x + scale, y, z);
    glVertex3f(x + scale, y + scale, z);
    glVertex3f(x, y + scale, z);
    glVertex3f(x, y, z);
    glEnd();

    glBegin(GL_LINE_STRIP);
    glVertex3f(x, y, z + scale);
    glVertex3f(x + scale, y, z + scale);
    glVertex3f(x + scale, y + scale, z + scale);
    glVertex3f(x, y + scale, z + scale);
    glVertex3f(x, y, z + scale);
    glEnd();

    glBegin(GL_LINES);
    glVertex3f(x, y, z);
    glVertex3f(x, y, z + scale);

    glVertex3f(x + scale, y, z);
    glVertex3f(x + scale, y, z + scale);

    glVertex3f(x + scale, y + scale, z);
    glVertex3f(x + scale, y + scale, z + scale);

    glVertex3f(x, y + scale, z);
    glVertex3f(x, y + scale, z + scale);
    glEnd();

    if(Light)
    {
        glEnable(GL_LIGHTING);
    }
}

void MarchingCubesGL::Resize(const GLsizei width, const GLsizei height)
{
    // constexpr GLfloat halfWorld = (1.4142135623730950488016887242097f / 2.0f);
    constexpr GLfloat halfWorld = 1.0f;

    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if(width <= height)
    {
        const GLfloat aspect = static_cast<GLfloat>(height) / static_cast<GLfloat>(width);
        glOrtho(
            -halfWorld, halfWorld,
            -halfWorld * aspect, halfWorld * aspect,
            -10 * halfWorld, 10 * halfWorld);
    }
    else
    {
        const GLfloat aspect = static_cast<GLfloat>(width) / static_cast<GLfloat>(height);
        glOrtho(
            -halfWorld * aspect, halfWorld * aspect,
            -halfWorld, halfWorld,
            -10 * halfWorld, 10 * halfWorld);
    }

    glMatrixMode(GL_MODELVIEW);
}

void MarchingCubesGL::Keyboard(const unsigned char key, int x, int y)
{
    const auto keyModifiers = glutGetModifiers();
    switch(key)
    {
        case 'e':
        {
            ShowSurface = !ShowSurface;
        }
        break;
        case 'l':
        {
            if(Light)
            {
                glDisable(GL_LIGHTING);    // use vertex colors
            }
            else
            {
                glEnable(GL_LIGHTING);    // use lit material color
            }

            Light = !Light;
        }
        break;
        case 'R':
        case 'r':
        {
            PolygonMode  = GL_FILL;
            RowSize      = 16;
            StepSize     = 1.0f / static_cast<GLfloat>(RowSize);
            SurfaceValue = 1.0f;
            Time         = 0.0f;
            Spin         = false;
            SlowSpin     = true;
            Move         = false;
            SlowMove     = true;
            Light        = true;
            ShowSurface  = true;
        }
        break;
        case 'S':
        case 's':
        {
            if(keyModifiers == GLUT_ACTIVE_SHIFT)
            {
                SampleNum--;
            }
            else
            {
                SampleNum++;
            }

            if(SampleNum > SampleCount)
            {
                SampleNum = 1;
            }

            if(SampleNum < 1)
            {
                SampleNum = SampleCount;
            }
            // clang-format off
            switch(SampleNum)
            {
                case 1:  { glutSetWindowTitle("Marching Cubes: Sample1");  Sample = Sample1;  } break;
                case 2:  { glutSetWindowTitle("Marching Cubes: Sample2");  Sample = Sample2;  } break;
                case 3:  { glutSetWindowTitle("Marching Cubes: Sample3");  Sample = Sample3;  } break;
                case 4:  { glutSetWindowTitle("Marching Cubes: Sample4");  Sample = Sample4;  } break;
                case 5:  { glutSetWindowTitle("Marching Cubes: Sample5");  Sample = Sample5;  } break;
                case 6:  { glutSetWindowTitle("Marching Cubes: Sample6");  Sample = Sample6;  } break;
                case 7:  { glutSetWindowTitle("Marching Cubes: Sample7");  Sample = Sample7;  } break;
                case 8:  { glutSetWindowTitle("Marching Cubes: Sample8");  Sample = Sample8;  } break;
                case 9:  { glutSetWindowTitle("Marching Cubes: Sample9");  Sample = Sample9;  } break;
                case 10: { glutSetWindowTitle("Marching Cubes: Sample10"); Sample = Sample10; } break;
                case 11: { glutSetWindowTitle("Marching Cubes: Sample11"); Sample = Sample11; } break;
                default: { glutSetWindowTitle("Marching Cubes: Sample1");  Sample = Sample1;  }
            }
            // clang-format on
        }
        break;
        case 'Q':
        case 'q':
        {
            exit(0);
        }
        case 'w':
        {
            if(PolygonMode == GL_LINE)
            {
                PolygonMode = GL_FILL;
            }
            else
            {
                PolygonMode = GL_LINE;
            }
            glPolygonMode(GL_FRONT_AND_BACK, PolygonMode);
        }
        break;
        case '+':
        case '=':
        {
            ++RowSize;
            StepSize = 1.0f / static_cast<float>(RowSize);
            // PrintLog();
        }
        break;
        case '_':
        case '-':
        {
            if(RowSize > 1)
            {
                --RowSize;
                StepSize = 1.0f / static_cast<float>(RowSize);
            }
            // PrintLog();
        }
        break;
        case '1':
        {
            BBoxInternal = !BBoxInternal;
        }
        break;
        case '2':
        {
            BBoxExtenal = !BBoxExtenal;
        }
        break;
        case '3':
        {
            PointsInternal = !PointsInternal;
        }
        break;
        case '4':
        {
            PointsInternalWhite = !PointsInternalWhite;
        }
        break;
        case '5':
        {
            PointsExtenal = !PointsExtenal;
        }
        break;
        case '6':
        {
            PointsExtenalWhite = !PointsExtenalWhite;
        }
    }
}

void MarchingCubesGL::Special(const int key, int x, int y)
{
    const auto keyModifiers = glutGetModifiers();
    switch(key)
    {
        case GLUT_KEY_PAGE_UP:
        {
            if(SurfaceValue < 1000.0 && SurfaceValue > 0.00000001)
            {
                SurfaceValue *= keyModifiers == GLUT_ACTIVE_SHIFT ? 1.01f : 1.1f;
            }
            // PrintLog();
        }
        break;
        case GLUT_KEY_PAGE_DOWN:
        {
            if(SurfaceValue > 0.0000001)
            {
                SurfaceValue /= keyModifiers == GLUT_ACTIVE_SHIFT ? 1.01f : 1.1f;
            }
            // PrintLog();
        }
        break;
        case GLUT_KEY_HOME:
        {
            Spin     = !Spin;
            SlowSpin = keyModifiers == GLUT_ACTIVE_SHIFT && Spin;
        }
        break;
        case GLUT_KEY_END:
        {
            Move     = !Move;
            SlowMove = keyModifiers == GLUT_ACTIVE_SHIFT && Move;
        }
        break;
    }
}
